#!/bin/sh
# the next line restarts using wish (must be wish4.0 or newer) \
exec wish "$0" "$@"

# gridworld demo
# version $Id: grid.tcl,v 1.3 1999/02/17 17:45:27 uwe Exp $

# auto_path and tclIndex not yet needed
# auto_mkindex . *.tcl
# set auto_path [concat . $auto_path]

catch {unset field}
eval destroy [winfo children .]

# menu bar widget -----------------------------------------------------------------

wm title      . "Grid World Demo"
wm iconname   . "Grid World"
wm iconbitmap . @./grid_demo.60x60.xbm
wm geometry   . 400x400-10-10

# initial field_strategie
set field_strategie	{Goal Random}

frame .menu \
    -relief raised \
    -bd 1

pack  .menu -side top      -fill x

button .menu.logo \
    -relief flat \
    -text "Grid World Demo" \
    -underline 0 \
    -fg IndianRed4 \
    -activeforeground IndianRed3 \
    -command "helpGridWorld"

button .menu.bitmap \
    -relief flat \
    -bitmap "@./grid_demo.15x15.xbm" \
    -fg IndianRed4 \
    -activeforeground IndianRed3 \
    -command "aboutGridWorld"

bind all <Alt-d> ".menu.logo invoke"

pack .menu.logo -side right
pack .menu.bitmap -side right

proc aboutGridWorld {} {

    regsub -all {\$} { Grid World Demo

Visualisation of path finding agents in a grid world

Author: Uwe Schmidt, uwe@fh-wedel.de
$Revision: 1.3 $ 
$Date: 1999/02/17 17:45:27 $ 
} "" text1

    regsub {..:..:..} $text1 "" text2

    tk_dialog .about "About Grid World Demo" $text2 "@./grid_demo.120x120.xbm" 0 ok

}

proc helpGridWorld {} {

    set text { Grid World Demo

Configuration

clear or resest field for configuration,
then press left mouse button over white and black fields for building or removing walls
press left mouse button over grey fields for positioning the yellow target field
press right mouse button over white fields for positioning the moving red field

use the Program menu entries for playing
use the Agents menu entries for selecting the search strategie
use the Option menu entries for setting grid size, maximum number of moves and speed
}

    tk_dialog .help "Help for Grid World Demo" $text "@./grid_demo.120x120.xbm" 0 ok

}

# program menue -------------------------------------------------------------------

set m  .menu.program
set mm $m.m

menubutton $m \
    -text "Program" \
    -menu $mm \
    -underline 0

menu $mm
$mm add command \
    -label "Clear" \
    -command clearField \
    -underline 0 \
    -accelerator Alt+c

$mm add command \
    -label "Reset" \
    -command resetField \
    -underline 0 \
    -accelerator Alt+r

$mm add command \
    -label "Start" \
    -command startField \
    -underline 0 \
    -accelerator Alt+s

$mm add command \
    -label "Stop" \
    -command stopField \
    -underline 3 \
    -accelerator Alt+p

$mm add command \
    -label "Continue" \
    -command restartField \
    -underline 1 \
    -accelerator Alt+o

$mm add separator

$mm add command \
    -label "Quit" \
    -command quitProgram \
    -underline 0 \
    -accelerator Alt+q

# agent menu ----------------------------------------------------------------------

set m .menu.agents
set mm $m.m

menubutton $m \
    -text Agents \
    -menu $mm \
    -underline 0

menu $mm
$mm add radio \
    -label "random walk" \
    -variable field_strategie \
    -value {
	Random
    }

$mm add radio \
    -label "random walk 1 step preview" \
    -variable field_strategie \
    -value $field_strategie

$mm add radio \
    -label "random walk 1 step preview, look for unvisited fields" \
    -variable field_strategie \
    -value {
	Goal
	RandomNew
	Random
    }

$mm add radio \
    -label "directed walk 1 step preview, look for unvisited fields" \
    -variable field_strategie \
    -value {
	Goal
	SameDirectionNew
	RandomNew
	Random
    }

$mm add radio \
    -label "goal directed walk 1 step preview, look for unvisited fields" \
    -variable field_strategie \
    -value {
	Goal
	GoalDirectionNew
	SameDirectionNew
	RandomNew
	Random
    }

$mm add radio \
    -label "along the walls, look for unvisited fields" \
    -variable field_strategie \
    -value {
	Goal
	AlongTheWallNewSameDirection
	AlongTheWallNew
	AlongTheWallNoDouble
	GoalDirectionNew
	SameDirectionNew
	RandomNew
	Random
    }


# options menu --------------------------------------------------------------------

set m  .menu.options
set mm $m.m

menubutton $m \
	-text Options \
	-menu $mm \
	-underline 0
menu $mm
$mm add cascade \
    -label "Grid size" \
    -menu $mm.dim

$mm add cascade \
    -label "delay" \
    -menu $mm.speed

$mm add cascade \
    -label "\# steps" \
    -menu $mm.steps

set mm .menu.options.m.dim
menu $mm

foreach i {{5 5} {10 10} {10 15} {20 20} {20 30} {50 50} {50 75}} {
    $mm add radio \
	-label "[lindex $i 0]x[lindex $i 1]" \
	-variable field_dimensions \
	-value  $i
}

set mm .menu.options.m.speed
menu $mm

foreach i {10 20 50 100 200 500 1000} {
    $mm add radio \
	-label $i \
	-variable field_delay \
	-value $i
}

set mm .menu.options.m.steps
menu $mm

foreach i {1 2 10 50 100 200 500 1000 2000 5000 10000} {
    $mm add radio \
	-label $i \
	-variable field_maxSteps \
	-value $i
}

$mm add radio \
    -label "unlimited" \
    -variable field(maxStep) \
    -value [expr 0x7fffffff]

unset m mm

pack .menu.program .menu.agents .menu.options -side left

# status line widget --------------------------------------------------------------

frame .statusline \
	-relief raised \
	-borderwidth 1

label .statusline.status \
    -textvariable field(message) \
    -anchor w \
    -fg IndianRed4

pack .statusline -fill x -side bottom

foreach i {status} {
    pack .statusline.$i \
	    -expand yes \
	    -side left \
	    -fill x
}

frame .grid \
	-border 1 \
	-relief sunken

pack .grid \
	-side top \
	-fill both \
	-expand yes

# the field object -------------------------------------------------------------------------

set field(message) ""

array set field {
    fieldBorder	grey
    fieldGoal	yellow
    fieldCurr	red
    fieldWall	black
    fieldFree	white
    fieldVisit	\#e9e9e9
}

proc setFieldStatus status {
    global field_status

    puts "set field_status $status"
    set field_status $status
}

proc isInitializable {} {
    global field_status

    return [expr [lsearch {empty finished stopped initialized} $field_status] != -1]
}

proc isConfigurable {} {
    global field_status

    return [expr [lsearch {initialized} $field_status] != -1]
}

proc isStartable {} {
    global field_status

    return [expr [lsearch {initialized} $field_status] != -1]
}

proc isRunning {} {
    global field_status

    return [expr [lsearch {running} $field_status] != -1]
}

proc isStopped {} {
    global field_status

    return [expr [lsearch {stopped} $field_status] != -1]
}

proc isRestartable {} {
    global field_status

    return [expr [lsearch {stopped initialized} $field_status] != -1]
}

proc isFinished {} {
    global field_status

    return [expr [lsearch {stopped finished} $field_status] != -1]
}

proc saveStatus {} {
    global field savedStatus

    # puts "status saved"

    set savedStatus [array get field]
    return
}

proc restoreStatus {} {
    global field savedStatus

    array set field $savedStatus

    # puts "status restored"
}

proc startField {} {

    if {! [isStartable]} return

    global field field_dimensions

    restoreStatus

    set field(steps) 0

    # set successor list for all fields to empty list
    # set visited to free

    set dim1 [expr [lindex $field_dimensions 0] + 2]
    set dim2 [expr [lindex $field_dimensions 1] + 2]

    for {set i 0} {$i < $dim1} {incr i} {
	for {set j 0} {$j < $dim2} {incr j} {
	    set field(succ,$i,$j) {}

	    if {"$field($i,$j)" == "fieldVisit"} {
		set field($i,$j) "fieldFree"
	    }
	}
    }

    # dummy move to initialize history
    set field(Curr) $field(Curr)

    restartField
}

proc restartField {} {

    if {! [isRestartable]} return

    setFieldStatus	"running"
    
    walk
}

proc stopField {} {

    if {! [isRunning]} return

    global field

    setFieldStatus	"stopped"
    set field(message)	"walking stopped"
}

proc resetField {} {

    if {! [isInitializable]} return

    global field field_status

    # restore initial state
    restoreStatus

    set field(Goal)	$field(Goal)
    setFieldStatus	"initialized"
    set field(message)	"grid reset"
}

proc clearField {} {

    if {! [isInitializable]} return

    global field field_dimensions

    set field(message)        "initializing grid"

    # unset all fields
    foreach i [ array names field {[0-9]*,[0-9]*} ] {
	unset field($i)
    }

    # unset move history
    foreach i [ array names field {succ,[0-9]*,[0-9]*} ] {
	unset field($i)
    }

    foreach w [winfo children .grid] {
	destroy $w
    }

    set dim1 [expr [lindex $field_dimensions 0] + 2]
    set dim2 [expr [lindex $field_dimensions 1] + 2]

    set relWidth  [expr double(1) / double($dim1)]
    set relHeight [expr double(1) / double($dim2)]

    for {set i 0} {$i < $dim1} {incr i} {
	for {set j 0} {$j < $dim2} {incr j} {

	    # mark field as free
	    set field($i,$j) fieldFree

	    # set successor move list to empty
	    set field(succ,$i,$j) {}
	    trace variable field($i,$j) w traceFieldIndex

	    set w .grid.$i,$j
	    label $w \
		-bd 1 \
		-relief raised \
		-highlightthickness 0 \
		-bg $field($field($i,$j))

	    bind $w <1> "pushButton $i,$j 1"
	    bind $w <3> "pushButton $i,$j 3"

	    set x [expr double($i) / double($dim1)]
	    set y [expr double($j) / double($dim2)]

	    place $w -in .grid -relx $x -rely $y -relwidth $relWidth -relheight $relHeight
	}
    }

    for {set i 0} {$i < $dim1} {incr i} {
	set field($i,0)                fieldBorder
	set field($i,[expr $dim2 - 1]) fieldBorder
    }

    for {set j 0} {$j < $dim2} {incr j} {
	set field(0,$j)                fieldBorder
	set field([expr $dim1 - 1],$j) fieldBorder
    }

    set xGoal [expr $dim1 - 1]
    set yGoal [expr $dim2 - 2]

    catch {unset field(Goal)}

    set field(Goal)     "$xGoal,$yGoal"
    set field(lastGoal) $field(Goal)
    trace variable field(Goal) w "traceGoal"
    set field(Goal)     $field(Goal)

    catch {unset field(Curr)}

    set field(Curr)     "1,1"
    set field(lastCurr) $field(Curr)
    trace variable field(Curr) w "traceCurr"
    set field(Curr)     $field(Curr)

    # save initial state
    saveStatus

    setFieldStatus	"initialized"
    set field(message)	"grid initialized"
}

proc traceDimensions {var ix s} {
    setFieldStatus "empty"
    clearField
}

proc traceGoal {var ix s} {
    global field

    set field($field(lastGoal)) fieldBorder
    set field($field(Goal))     fieldGoal
    set field(lastGoal)         $field(Goal)

}

proc sub {ix1 ix0} {
    set xy1 [split $ix1 ,]
    set xy0 [split $ix0 ,]

    return "[expr [lindex $xy1 0] - [lindex $xy0 0]],[expr [lindex $xy1 1] - [lindex $xy0 1]]"
}

proc add {ix1 ix0} {
    set xy1 [split $ix1 ,]
    set xy0 [split $ix0 ,]

    return "[expr [lindex $xy1 0] + [lindex $xy0 0]],[expr [lindex $xy1 1] + [lindex $xy0 1]]"
}

proc signum ix {
    set xy [split $ix ,]
    set x  [lindex $xy 0]
    set y  [lindex $xy 1]
    
    set x1 [expr $x > 0 ? 1 : $x < 0 ? -1 : 0]
    set y1 [expr $y > 0 ? 1 : $y < 0 ? -1 : 0]

    return "${x1},${y1}"
}

proc traceCurr {var ix s} {
    global field

    set old $field(lastCurr)
    set new $field(Curr)

    # extend move history
    lappend field(succ,$old)	$new
    set field(speed)		[sub $new $old]

    set field($old)		fieldVisit
    set field($new)		fieldCurr
    set field(lastCurr)		$new
}

proc traceFieldIndex {var ix s} {
    global field

    .grid.$ix configure \
	-background          $field($field($ix))
}

proc pushButton {ix b} {
    global field

    eval push$field($ix)Button $ix $b
}

proc pushfieldFreeButton {ix b} {
    global field

    if {[isConfigurable]} {
	if {$b == 1} {
	    set field($ix)	fieldWall
	} else {
	    set lastCurr		$field(Curr)
	    set field(Curr)		$ix
	    set field($lastCurr)	fieldFree
	}
	saveStatus
	set field(message)	"brick set"
    }
}

proc pushfieldWallButton {ix b} {
    global field field_status

    if {[isConfigurable]} {
	set field($ix) 	fieldFree
	saveStatus
	set field(message)	"brick removed"
    }
}

proc pushfieldBorderButton {ix b} {
    global field

    if {[isConfigurable]} {
	set field(Goal) 	$ix
	saveStatus
	set field(message)	"goal moved"
    }
}

proc pushfieldGoalButton {ix b} {
    # puts "pushyellowButton $ix"
}

proc pushfieldCurrButton {ix b} {
    # puts "pushredButton $ix"
}

proc pushfieldVisitButton {ix b} {
    # pushfieldFreeButton $ix $b
}
# ------------------------------------------------------------

proc walk {} {
    global field field_status field_maxSteps field_delay

    while {[isRunning]} {
	
	set mv [nextMove]

	puts "walk: $mv"

	if {"$mv" == "found"} {
	    setFieldStatus	"finished"
	    set ss ""
	    if {$field(steps) != 1} {set ss s}
	    set field(message) "finished: goal found in $field(steps) step$ss"
	} elseif {"$mv" == "max"} {
	    setFieldStatus	"stopped"
	    set field(message) "I give up: goal not found in $field_maxSteps steps"
	} elseif {"$mv" == "dead"} {
	    setFieldStatus	"finished"
	    set field(message) "I give up: goal not found"
	} else {
	    set field(message) "walking: $field(steps). step"
	    set field(Curr) $mv

	    update
	    after $field_delay
	}
    }
}

# ------------------------------------------------------------

proc allMoves ix {
    set xy [split $ix ,]
    set x  [lindex $xy 0]
    set y  [lindex $xy 1]

    set x-1 [expr $x - 1]
    set x+1 [expr $x + 1]

    set y-1 [expr $y - 1]
    set y+1 [expr $y + 1]

    set res {}
    for {set i ${x-1}} {$i <= ${x+1}} {incr i} {
	for {set j ${y-1}} {$j <= ${y+1}} {incr j} {
	    set ix ${i},${j}
	    if {$i != $x || $j != $y} {
		lappend res [join [list $i $j] ,]
	    }
	}
    }
    return $res
}

proc noWallMoves ixList {
    global field

    set res {}
    foreach ix $ixList {
	if {"$field($ix)" != "fieldBorder" && "$field($ix)" != "fieldWall"} {
	    lappend res $ix
	}
    }
    return $res
}

proc noVisitedMoves ixList {
    global field

    set res {}
    foreach ix $ixList {
	if {[llength $field(succ,$ix)] == 0} {
	    lappend res $ix
	}
    }
    return $res
}

proc noDoubleMoves ixList {
    global field

    set res {}
    foreach ix $ixList {
	if {[lsearch $field(succ,$field(Curr)) $ix] == -1} {
	    lappend res $ix
	}
    }
    return $res
}

proc alongTheWallMoves ixList {
    set res {}

    foreach ix $ixList {
	set mvl1 [allMoves $ix]
	set mvl2 [noWallMoves $mvl1]
	if {[llength $mvl1] > [llength $mvl2]} {
	    lappend res $ix
	}
    }
    return $res
}

proc alongTheWallNotVisitedMoves ixList {
    alongTheWallMoves [noVisitedMoves $ixList]
}

proc alongTheWallNoDoubleMoves ixList {
    alongTheWallMoves [noDoubleMoves $ixList]
}

# ------------------------------------------------------------

proc nextMoveGoal ixList {
    global field

    return [lindex $ixList [lsearch $ixList $field(Goal)]]
}

proc random {limit} {
    return [expr ([clock clicks] / 13) % $limit]
}

proc nextMoveRandom ixList {
    set l [llength $ixList]
    if {$l == 0} {
	return 
    }
    return [lindex $ixList [random $l]]
}

proc nextMoveRandomNew ixList {
    nextMoveRandom [noVisitedMoves $ixList]
}

proc nextMoveSameDirection ixList {
    global field

    set ix [add $field(Curr) $field(speed)]

    set res {}
    if {[lsearch $ixList $ix] != -1} {
	set res $ix
    }
    return $res
}

proc nextMoveSameDirectionNew ixList {
    nextMoveSameDirection [noVisitedMoves $ixList]
}

proc nextMoveGoalDirection ixList {
    global field

    set dir [signum [sub $field(Goal) $field(Curr)]]
    set ix  [add $field(Curr) $dir]
	     
    set res {}
    if {[lsearch $ixList $ix] != -1} {
	set res $ix
    }
    return $res
}

proc nextMoveGoalDirectionNew ixList {
    nextMoveGoalDirection [noVisitedMoves $ixList]
}

proc nextMoveAlongTheWall ixList {
    nextMoveRandom [alongTheWallMoves $ixList]
}

proc nextMoveAlongTheWallNoDouble ixList {
    nextMoveRandom [alongTheWallNoDoubleMoves $ixList]
}

proc nextMoveAlongTheWallSameDirection ixList {
    nextMoveSameDirection [alongTheWallMoves $ixList]
}

proc nextMoveAlongTheWallNew ixList {
    nextMoveRandom [alongTheWallNotVisitedMoves $ixList]
}

proc nextMoveAlongTheWallNewSameDirection ixList {
    nextMoveSameDirection [alongTheWallNotVisitedMoves $ixList]
}


proc nextMove {} {
    global next field field_maxSteps field_strategie

    if {"$field(Curr)" == "$field(Goal)"} {
	return found
    }
    if {$field(steps) >= $field_maxSteps} {
	return max
    }

    set mvl [noWallMoves [allMoves $field(Curr)]]

    foreach s $field_strategie {
	set ix [nextMove$s $mvl]
	if {"$ix" != ""} {
	    incr field(steps)
	    return $ix
	}
    }

    return "dead"
}

# quitProgram ------------------------------------------------------------

# stops sort process
# and terminates application

proc quitProgram {} {
    setFieldStatus stopped
    update

    destroy .
}

# set trace variable functions -------------------------------------------

set field_status	empty
set field_dimensions	{10 10}
set field_delay		1000
set field_maxSteps	10

trace variable field_dimensions		w traceDimensions

# initialize application -------------------------------------------------

clearField

